import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np

from .modules import Transformer, PairwiseTransformer


class EmbeddingTransformer(nn.Module):
    def __init__(
        self, n_feats, n_layers, n_heads, d_model, d_ff, pos_enc=True, activation=F.relu
    ):
        super(EmbeddingTransformer, self).__init__()

        self.d_model = d_model
        self.lin_proj = nn.Linear(n_feats, self.d_model, bias=False)

        self.transformer = Transformer(
            n_layers, n_heads, self.d_model, self.d_model, d_ff, activation, gate=False
        )
        self.pos_enc = pos_enc

    def _pos_enc(self, dim, max_len):
        pe = torch.zeros(max_len, dim)
        pos = torch.arange(0, max_len, 1).unsqueeze(1)
        k = torch.exp(-np.log(10000) * torch.arange(0, dim, 2) / dim)
        pe[:, 0::2] = torch.sin(pos * k)
        pe[:, 1::2] = torch.cos(pos * k)
        return pe

    def forward(self, x, mask):
        """
        x : bsz x (max_n_interactions + 1) x max_n_bins x n_mark
        mask : bsz x (max_n_interactions + 1) x 1 x max_n_bins x max_n_bins
        """
        bsz, max_n_bins, n_mark = mask.size(0), mask.size(4), x.size(3)
        mask = mask.view(-1, 1, max_n_bins, max_n_bins)
        # --> mask : (bsz x (max_n_interactions + 1)) x 1 x max_n_bins x max_n_bins

        x = x.view(-1, max_n_bins, n_mark)
        # --> x: (bsz x (max_n_interactions + 1)) x max_n_bins x n_maxk
        x = self.lin_proj(x)
        # --> x: (bsz x (max_n_interactions + 1)) x max_n_bins x d_model

        # Embed and linearly project each genomic region independently.
        # Add positional embedding.
        if self.pos_enc:
            x = (
                x
                + self._pos_enc(dim=self.d_model, max_len=max_n_bins)[:, : self.d_model]
                .unsqueeze(0)
                .cuda()
            )
        # --> x : (bsz x (max_n_interactions + 1)) x max_n_bins x d_model
        x = self.transformer(x, mask)
        # --> x : (bsz x (max_n_interactions + 1)) x max_n_bins x d_model
        x = x.view(bsz, -1, max_n_bins, self.d_model)

        return x, x[:, :, max_n_bins // 2]


class PairwiseInteractionTransformer(nn.Module):
    def __init__(
        self,
        n_feats_p,
        n_feats_pcre,
        n_layers,
        n_heads,
        d_model,
        d_ff,
        pos_enc=True,
        res_p=0.0,
        att_p=0.0,
        activation=F.relu,
    ):
        super(PairwiseInteractionTransformer, self).__init__()

        self.d_model = d_model

        self.ln = nn.LayerNorm(self.d_model)
        self.lin_proj_p = nn.Linear(n_feats_p, self.d_model, bias=False)
        self.lin_proj_pcre = nn.Linear(n_feats_pcre, self.d_model, bias=False)

        self.transformer = PairwiseTransformer(
            n_layers,
            n_heads,
            self.d_model,
            self.d_model,
            d_ff,
            res_p,
            att_p,
            activation,
            gate=False,
        )
        self.pos_enc = pos_enc

    def _pos_enc(self, dim, max_len):
        pe = torch.zeros(max_len, dim)
        pos = torch.arange(0, max_len, 1).unsqueeze(1)
        k = torch.exp(-np.log(10000) * torch.arange(0, dim, 2) / dim)
        pe[:, 0::2] = torch.sin(pos * k)
        pe[:, 1::2] = torch.cos(pos * k)
        return pe

    def forward(self, x_p, x_pcre, mask):
        """
        x : bsz x (max_n_interactions + 1) x max_n_bins x n_mark
        mask : bsz x (max_n_interactions + 1) x 1 x max_n_bins x max_n_bins
        """
        bsz, max_n_bins, n_mark = mask.size(0), mask.size(4), x_pcre.size(3)
        mask = mask.view(-1, 1, max_n_bins, max_n_bins)
        # --> mask : (bsz x (max_n_interactions + 1)) x 1 x max_n_bins x max_n_bins

        x_p = x_p.expand(-1, x_pcre.size(1), -1, -1)
        x_p = x_p.reshape(-1, x_p.size(2), x_p.size(3))
        x_pcre = x_pcre.view(-1, max_n_bins, n_mark)
        # --> x: (bsz x (max_n_interactions + 1)) x max_n_bins x n_maxk
        x_p = self.lin_proj_p(x_p)
        x_pcre = self.lin_proj_pcre(x_pcre)
        # --> x: (bsz x (max_n_interactions + 1)) x max_n_bins x d_model

        # Embed and linearly project each genomic region independently.
        # Add positional embedding.
        if self.pos_enc:
            x_pcre = (
                x_pcre
                + self._pos_enc(dim=self.d_model, max_len=max_n_bins)[:, : self.d_model]
                .unsqueeze(0)
                .cuda()
            )
        # --> x : (bsz x (max_n_interactions + 1)) x max_n_bins x d_model
        x = self.transformer(x_p, x_pcre, mask)
        # --> x : (bsz x (max_n_interactions + 1)) x max_n_bins x d_model

        x = x.view(bsz, -1, max_n_bins, self.d_model)
        # --> x : bsz x (max_n_interactions + 1) x max_n_bins x d_model

        x = x[:, :, max_n_bins // 2]  # Take the genomic bin representation at center.
        return x


class RegulationTransformer(nn.Module):
    def __init__(
        self, n_layers, n_heads, d_emb, d_model, d_ff, pos_enc=True, activation=F.relu
    ):
        super(RegulationTransformer, self).__init__()

        self.transformer = Transformer(
            n_layers, n_heads, d_emb, d_model, d_ff, gate=True
        )

    def forward(self, x, mask, bias):
        return self.transformer(x, mask, bias)


class Chromoformer(nn.Module):
    def __init__(
        self,
        n_feats=7,
        embed_n_layers=1,
        embed_n_heads=2,
        embed_d_model=128,
        embed_d_ff=128,
        pw_int_n_layers=2,
        pw_int_n_heads=2,
        pw_int_d_model=128,
        pw_int_d_ff=256,
        reg_n_layers=6,
        reg_n_heads=8,
        reg_d_model=256,
        reg_d_ff=256,
        head_n_feats=128,
        seed=42,
    ):
        super(Chromoformer, self).__init__()
        torch.manual_seed(seed)
        self.embed2000 = EmbeddingTransformer(
            n_feats, embed_n_layers, embed_n_heads, embed_d_model, embed_d_ff
        )
        self.embed500 = EmbeddingTransformer(
            n_feats, embed_n_layers, embed_n_heads, embed_d_model, embed_d_ff
        )
        self.embed100 = EmbeddingTransformer(
            n_feats, embed_n_layers, embed_n_heads, embed_d_model, embed_d_ff
        )

        self.pw_int2000 = PairwiseInteractionTransformer(
            embed_d_model,
            n_feats,
            pw_int_n_layers,
            pw_int_n_heads,
            pw_int_d_model,
            pw_int_d_ff,
        )
        self.pw_int500 = PairwiseInteractionTransformer(
            embed_d_model,
            n_feats,
            pw_int_n_layers,
            pw_int_n_heads,
            pw_int_d_model,
            pw_int_d_ff,
        )
        self.pw_int100 = PairwiseInteractionTransformer(
            embed_d_model,
            n_feats,
            pw_int_n_layers,
            pw_int_n_heads,
            pw_int_d_model,
            pw_int_d_ff,
        )

        self.reg2000 = RegulationTransformer(
            reg_n_layers, reg_n_heads, embed_d_model, reg_d_model, reg_d_ff
        )
        self.reg500 = RegulationTransformer(
            reg_n_layers, reg_n_heads, embed_d_model, reg_d_model, reg_d_ff
        )
        self.reg100 = RegulationTransformer(
            reg_n_layers, reg_n_heads, embed_d_model, reg_d_model, reg_d_ff
        )

        self.fc_head = nn.Sequential(
            nn.Linear(embed_d_model * 3, head_n_feats),
            nn.ReLU(),
            nn.Linear(head_n_feats, 2),
        )

    def forward(
        self,
        x_p_2000,
        pad_mask_p_2000,
        x_pcre_2000,
        pad_mask_pcre_2000,
        interaction_mask_2000,
        x_p_500,
        pad_mask_p_500,
        x_pcre_500,
        pad_mask_pcre_500,
        interaction_mask_500,
        x_p_100,
        pad_mask_p_100,
        x_pcre_100,
        pad_mask_pcre_100,
        interaction_mask_100,
        interaction_freq,
    ):
        p_emb_2000_total, p_emb_2000 = self.embed2000(x_p_2000, pad_mask_p_2000)
        p_emb_500_total, p_emb_500 = self.embed500(x_p_500, pad_mask_p_500)
        p_emb_100_total, p_emb_100 = self.embed100(x_p_100, pad_mask_p_100)

        pw_int_emb_2000 = self.pw_int2000(
            p_emb_2000_total, x_pcre_2000, pad_mask_pcre_2000
        )
        pw_int_emb_500 = self.pw_int500(p_emb_500_total, x_pcre_500, pad_mask_pcre_500)
        pw_int_emb_100 = self.pw_int100(p_emb_100_total, x_pcre_100, pad_mask_pcre_100)

        x_2000 = torch.cat([p_emb_2000, pw_int_emb_2000], axis=1)
        x_500 = torch.cat([p_emb_500, pw_int_emb_500], axis=1)
        x_100 = torch.cat([p_emb_100, pw_int_emb_100], axis=1)

        emb = torch.cat([x_2000[:, 0], x_500[:, 0], x_100[:, 0]], axis=1)

        x_2000 = self.reg2000(x_2000, interaction_mask_2000, bias=interaction_freq)
        x_500 = self.reg500(x_500, interaction_mask_500, bias=interaction_freq)
        x_100 = self.reg100(x_100, interaction_mask_100, bias=interaction_freq)

        x = (
            torch.cat([x_2000[:, 0], x_500[:, 0], x_100[:, 0]], axis=1) + emb
        )  # Take representation of the target promoter.
        return self.fc_head(x)


class ChromoformerBase(nn.Module):
    def __init__(
        self,
        n_feats=7,
        d_emb=128,
        d_head=128,
        embed_kws={
            "n_layers": 1,
            "n_heads": 2,
            "d_model": 128,
            "d_ff": 128,
        },
        pairwise_interaction_kws={
            "n_layers": 2,
            "n_heads": 2,
            "d_model": 128,
            "d_ff": 256,
        },
        regulation_kws={
            "n_layers": 6,
            "n_heads": 8,
            "d_model": 256,
            "d_ff": 256,
        },
        binsizes=[2000, 500, 100],
        seed=42,
    ):
        super(ChromoformerBase, self).__init__()
        torch.manual_seed(seed)

        # Update arguments for each transformer layer.
        embed_kws["n_feats"] = n_feats
        embed_kws["d_model"] = d_emb
        pairwise_interaction_kws["n_feats_p"] = embed_kws["d_model"]
        pairwise_interaction_kws["n_feats_pcre"] = n_feats
        regulation_kws["d_emb"] = d_emb

        self.binsizes = binsizes
        self.embed = nn.ModuleDict(
            {str(binsize): EmbeddingTransformer(**embed_kws) for binsize in binsizes}
        )
        self.pairwise_interaction = nn.ModuleDict(
            {
                str(binsize): PairwiseInteractionTransformer(**pairwise_interaction_kws)
                for binsize in binsizes
            }
        )
        self.regulation = nn.ModuleDict(
            {
                str(binsize): RegulationTransformer(**regulation_kws)
                for binsize in binsizes
            }
        )
        self.fc_head = nn.Sequential(
            nn.Linear(d_emb * 3, d_head),
            nn.ReLU(),
            nn.Linear(d_head, 2),
        )

    def forward(
        self,
        promoter_feats,
        promoter_pad_masks,
        pcre_feats,
        pcre_pad_masks,
        interaction_masks,
        interaction_freq,
    ):
        promoter_embeddings_full, promoter_embeddings_tss = {}, {}
        for binsize in self.binsizes:
            layer = self.embed[str(binsize)]
            p_emb_full, p_emb = layer(
                promoter_feats[binsize], promoter_pad_masks[binsize]
            )
            promoter_embeddings_full[binsize] = p_emb_full
            promoter_embeddings_tss[binsize] = p_emb

        pairwise_interaction_embeddings = {}
        for binsize in self.binsizes:
            layer = self.pairwise_interaction[str(binsize)]
            pairwise_interaction_embeddings[binsize] = layer(
                promoter_embeddings_full[binsize],
                pcre_feats[binsize],
                pcre_pad_masks[binsize],
            )

        x_in = {
            binsize: torch.cat(
                [
                    promoter_embeddings_tss[binsize],
                    pairwise_interaction_embeddings[binsize],
                ],
                axis=1,
            )
            for binsize in self.binsizes
        }

        x_out = {}
        for binsize in self.binsizes:
            layer = self.regulation[str(binsize)]
            x_out[binsize] = layer(
                x_in[binsize], interaction_masks[binsize], bias=interaction_freq
            )

        x = torch.cat([x_out[binsize][:, 0] for binsize in self.binsizes], axis=1)
        x += torch.cat([x_in[binsize][:, 0] for binsize in self.binsizes], axis=1)

        return self.fc_head(x)


ChromoformerClassifier = ChromoformerBase


class ChromoformerRegressor(ChromoformerBase):
    def __init__(
        self,
        n_feats=7,
        d_emb=128,
        d_head=128,
        embed_kws={
            "n_layers": 1,
            "n_heads": 2,
            "d_model": 128,
            "d_ff": 128,
        },
        pairwise_interaction_kws={
            "n_layers": 2,
            "n_heads": 2,
            "d_model": 128,
            "d_ff": 256,
        },
        regulation_kws={
            "n_layers": 6,
            "n_heads": 8,
            "d_model": 256,
            "d_ff": 256,
        },
        binsizes=[2000, 500, 100],
        seed=42,
    ):
        super(ChromoformerRegressor, self).__init__(
            n_feats,
            d_emb,
            d_head,
            embed_kws,
            pairwise_interaction_kws,
            regulation_kws,
            binsizes,
            seed,
        )

        self.fc_head = nn.Sequential(
            nn.Linear(d_emb * 3, d_head),
            nn.ReLU(),
            nn.Linear(d_head, 1),
        )


if __name__ == "__main__":
    # import data
    model_orig = Chromoformer().cuda()
    model_classifier = ChromoformerClassifier().cuda()
    model_regressor = ChromoformerRegressor().cuda()

    # Dummy data.
    bsz = 8
    i_max = 8

    x_p_2000, x_p_500, x_p_100 = (
        torch.randn([bsz, 1, 20, 7]),
        torch.randn([bsz, 1, 80, 7]),
        torch.randn([bsz, 1, 400, 7]),
    )
    x_pcre_2000, x_pcre_500, x_pcre_100 = (
        torch.randn([bsz, i_max, 20, 7]),
        torch.randn([bsz, i_max, 80, 7]),
        torch.randn([bsz, i_max, 400, 7]),
    )

    pad_mask_p_2000, pad_mask_p_500, pad_mask_p_100 = (
        torch.randn([bsz, 1, 1, 20, 20]).bool(),
        torch.randn([bsz, 1, 1, 80, 80]).bool(),
        torch.randn([bsz, 1, 1, 400, 400]).bool(),
    )
    pad_mask_pcre_2000, pad_mask_pcre_500, pad_mask_pcre_100 = (
        torch.randn([bsz, i_max, 1, 20, 20]).bool(),
        torch.randn([bsz, i_max, 1, 80, 80]).bool(),
        torch.randn([bsz, i_max, 1, 400, 400]).bool(),
    )

    interaction_mask_2000, interaction_mask_500, interaction_mask_100 = (
        torch.randn([bsz, 1, 1 + i_max, 1 + i_max]).bool(),
        torch.randn([bsz, 1, 1 + i_max, 1 + i_max]).bool(),
        torch.randn([bsz, 1, 1 + i_max, 1 + i_max]).bool(),
    )
    interaction_freq = torch.randn([bsz, 1 + i_max, 1 + i_max])

    x_p_2000, x_p_500, x_p_100 = x_p_2000.cuda(), x_p_500.cuda(), x_p_100.cuda()
    x_pcre_2000, x_pcre_500, x_pcre_100 = (
        x_pcre_2000.cuda(),
        x_pcre_500.cuda(),
        x_pcre_100.cuda(),
    )

    pad_mask_p_2000, pad_mask_p_500, pad_mask_p_100 = (
        pad_mask_p_2000.cuda(),
        pad_mask_p_500.cuda(),
        pad_mask_p_100.cuda(),
    )
    pad_mask_pcre_2000, pad_mask_pcre_500, pad_mask_pcre_100 = (
        pad_mask_pcre_2000.cuda(),
        pad_mask_pcre_500.cuda(),
        pad_mask_pcre_100.cuda(),
    )

    interaction_mask_2000, interaction_mask_500, interaction_mask_100 = (
        interaction_mask_2000.cuda(),
        interaction_mask_500.cuda(),
        interaction_mask_100.cuda(),
    )
    interaction_freq = interaction_freq.cuda()

    out_orig = model_orig(
        x_p_2000,
        pad_mask_p_2000,
        x_pcre_2000,
        pad_mask_pcre_2000,
        interaction_mask_2000,
        x_p_500,
        pad_mask_p_500,
        x_pcre_500,
        pad_mask_pcre_500,
        interaction_mask_500,
        x_p_100,
        pad_mask_p_100,
        x_pcre_100,
        pad_mask_pcre_100,
        interaction_mask_100,
        interaction_freq,
    )

    promoter_feats = {
        2000: x_p_2000,
        500: x_p_500,
        100: x_p_100,
    }
    promoter_pad_masks = {
        2000: pad_mask_p_2000,
        500: pad_mask_p_500,
        100: pad_mask_p_100,
    }
    pcre_feats = {
        2000: x_pcre_2000,
        500: x_pcre_500,
        100: x_pcre_100,
    }
    pcre_pad_masks = {
        2000: pad_mask_pcre_2000,
        500: pad_mask_pcre_500,
        100: pad_mask_pcre_100,
    }
    interaction_masks = {
        2000: interaction_mask_2000,
        500: interaction_mask_500,
        100: interaction_mask_100,
    }

    out_classifier = model_classifier(
        promoter_feats,
        promoter_pad_masks,
        pcre_feats,
        pcre_pad_masks,
        interaction_masks,
        interaction_freq,
    )

    out_regressor = model_regressor(
        promoter_feats,
        promoter_pad_masks,
        pcre_feats,
        pcre_pad_masks,
        interaction_masks,
        interaction_freq,
    )

    print(out_orig.sum())
    print(out_orig.shape)
    # -3.1917, [8, 2]

    print(out_classifier.sum())
    print(out_classifier.shape)
    # -3.1917, [8, 2]

    print(out_regressor.sum())
    print(out_regressor.shape)
    # -0.1900, [8, 1]

    # ckpt = torch.load("test.pt")
    # model_classifier.load_state_dict(ckpt["net"])

    ckpt = torch.load(
        "converted/E003/chromoformer-reg-reproduction-E003-conf1-fold1.pt"
    )
    model_regressor.load_state_dict(ckpt["net"])



import argparse
import torch
import torch.nn as nn
import os
import pandas as pd
import wandb
import yaml

from tqdm import tqdm
from scipy import stats
from sklearn import metrics

from .data import ChromoformerDataset
from .net import ChromoformerClassifier, ChromoformerRegressor
from .util import seed_everything


torch.autograd.set_detect_anomaly(True)


def get_lr(optimizer):
    for param_group in optimizer.param_groups:
        return param_group["lr"]


parser = argparse.ArgumentParser()
parser.add_argument("-o", "--output", required=True)
parser.add_argument("-c", "--config", required=True)
parser.add_argument("--exp-id", required=True)
parser.add_argument("-m", "--meta", required=True)
parser.add_argument("-d", "--npy-dir", required=True)
parser.add_argument("--fold", type=int, required=True)
parser.add_argument("--binsizes", nargs="+", default=[2000, 500, 100])
parser.add_argument("--regression", action="store_true", default=False)
parser.add_argument("--use-wandb", action="store_true", default=False)

args = parser.parse_args()

#
# Training setup.
#
if args.use_wandb is False:
    os.environ["WANDB_MODE"] = "disabled"

with open(args.config) as f:
    config = yaml.safe_load(f)

print(config)

config["exp_id"] = args.exp_id  # Override
exp_id = args.exp_id
seed = config["seed"]
num_epoch = config["num_epoch"]
lr = config["lr"]
bsz = config["bsz"]
gamma = config["gamma"]

i_max = config["i_max"]
w_prom = config["w_prom"]
w_max = config["w_max"]

n_feats = config["n_feats"]
d_emb = config["embed"]["d_model"]
embed_kws = config["embed"]
pairwise_interaction_kws = config["pairwise_interaction"]
regulation_kws = config["regulation"]

d_head = config["d_head"]

#
# Setup end.
#

seed_everything(seed)
wandb.init(project="chromoformer-refactoring", entity="dohlee", group=exp_id)
wandb.config.update(args)
wandb.config.update(config)

meta = (
    pd.read_csv(args.meta).sample(frac=1, random_state=seed).reset_index(drop=True)
)  # load and shuffle.

if args.regression and "expression" not in meta.columns:
    raise ValueError(
        "`expression` column is required for training ChromoformerRegression model."
    )

# Split genes into two sets (train/val).
genes = set(meta.gene_id.unique())
n_genes = len(genes)
print("Target genes:", len(genes))

qs = [
    meta[meta.split == 1].gene_id.tolist(),
    meta[meta.split == 2].gene_id.tolist(),
    meta[meta.split == 3].gene_id.tolist(),
    meta[meta.split == 4].gene_id.tolist(),
]

train_genes = (
    qs[(args.fold + 0) % 4] + qs[(args.fold + 1) % 4] + qs[(args.fold + 2) % 4]
)
val_genes = qs[(args.fold + 3) % 4]

wandb.config.update(
    {
        "n_train_genes": len(train_genes),
        "n_val_genes": len(val_genes),
    }
)

print(len(train_genes), len(val_genes))

train_dataset = ChromoformerDataset(
    args.meta,
    args.npy_dir,
    train_genes,
    n_feats,
    i_max,
    args.binsizes,
    w_prom,
    w_max,
    regression=args.regression,
)
val_dataset = ChromoformerDataset(
    args.meta,
    args.npy_dir,
    val_genes,
    n_feats,
    i_max,
    args.binsizes,
    w_prom,
    w_max,
    regression=args.regression,
)

train_loader = torch.utils.data.DataLoader(
    train_dataset, batch_size=bsz, num_workers=8, shuffle=True, drop_last=True
)
val_loader = torch.utils.data.DataLoader(val_dataset, batch_size=bsz, num_workers=8)

Model = ChromoformerRegressor if args.regression else ChromoformerClassifier
model = Model(
    n_feats,
    d_emb,
    d_head,
    embed_kws,
    pairwise_interaction_kws,
    regulation_kws,
    binsizes=args.binsizes,
    seed=42,
)

model.cuda()

criterion = nn.MSELoss() if args.regression else nn.CrossEntropyLoss()
optimizer = torch.optim.AdamW(model.parameters(), lr=float(config["lr"]))
scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=1, gamma=gamma)

optimizer.zero_grad()
optimizer.step()

for epoch in range(1, num_epoch):
    # Prepare train.
    bar = tqdm(enumerate(train_loader, 1), total=len(train_loader))
    running_loss = 0.0
    train_out, train_label = [], []

    # Train.
    model.train()
    for batch, d in bar:
        for k, v in d.items():
            if isinstance(v, dict):
                for _k, _v in v.items():
                    v[_k] = _v.cuda()
            else:
                d[k] = v.cuda()

        if args.regression:
            d["label"] = d["label"].view(-1, 1)

        optimizer.zero_grad()

        out = model(
            d["promoter_feats"],
            d["promoter_pad_masks"],
            d["pcre_feats"],
            d["pcre_pad_masks"],
            d["interaction_masks"],
            d["interaction_freq"],
        )

        loss = criterion(out, d["label"])

        loss.backward()
        optimizer.step()

        loss = loss.detach().cpu().item()
        running_loss += loss

        train_out.append(out.detach().cpu())
        train_label.append(d["label"].cpu())

        if batch % 10 == 0:
            batch_loss = running_loss / 10.0

            train_out, train_label = map(torch.cat, (train_out, train_label))

            if args.regression:
                train_pred = train_out.flatten()
                train_label = train_label.flatten()

                batch_r2 = metrics.r2_score(train_label, train_pred) * 100
                batch_r = stats.pearsonr(train_label, train_pred)[0] * 100

                bar.set_description(
                    f"E{epoch} {batch_loss:.4f}, lr={get_lr(optimizer)}, r2={batch_r2:.4f}, r={batch_r:.4f}"
                )

                wandb.log(
                    {
                        "train/loss": batch_loss,
                        "train/r2": batch_r2,
                        "train/r": batch_r,
                    }
                )
            else:
                train_score = train_out.softmax(axis=1)[:, 1]
                train_pred = train_out.argmax(axis=1)

                batch_acc = metrics.accuracy_score(train_label, train_pred) * 100
                batch_auc = metrics.roc_auc_score(train_label, train_score) * 100
                batch_ap = (
                    metrics.average_precision_score(train_label, train_score) * 100
                )

                bar.set_description(
                    f"E{epoch} {batch_loss:.4f}, lr={get_lr(optimizer)}, acc={batch_acc:.4f}, auc={batch_auc:.4f}, ap={batch_ap:.4f}"
                )

                wandb.log(
                    {
                        "train/loss": batch_loss,
                        "train/acc": batch_acc,
                        "train/auc": batch_auc,
                        "train/ap": batch_ap,
                    }
                )

            running_loss = 0.0
            train_out, train_label = [], []

    # Prepare validation.
    bar = tqdm(enumerate(val_loader, 1), total=len(val_loader))
    val_out, val_label = [], []

    # Validation.
    model.eval()
    with torch.no_grad():
        for batch, d in bar:
            for k, v in d.items():
                if isinstance(v, dict):
                    for _k, _v in v.items():
                        v[_k] = _v.cuda()
                else:
                    d[k] = v.cuda()

            out = model(
                d["promoter_feats"],
                d["promoter_pad_masks"],
                d["pcre_feats"],
                d["pcre_pad_masks"],
                d["interaction_masks"],
                d["interaction_freq"],
            )
            val_out.append(out.cpu())

            val_label.append(d["label"].cpu())

    val_out = torch.cat(val_out)
    val_label = torch.cat(val_label)

    val_loss = criterion(val_out, val_label)

    # Metrics.
    val_label = val_label.numpy()
    if args.regression:
        val_score = val_out.flatten().numpy()

        val_r2 = metrics.r2_score(val_label, val_score) * 100
        val_r = stats.pearsonr(val_label, val_score)[0] * 100

        print(f"Validation loss={val_loss:.4f}, r2={val_r2:.4f}, r={val_r:.4f}")
        wandb.log(
            {
                "val/loss": val_loss,
                "val/r2": val_r2,
                "val/r": val_r,
            }
        )
    else:
        val_score = val_out.softmax(axis=1)[:, 1].numpy()
        val_pred = val_out.argmax(axis=1).numpy()

        val_acc = metrics.accuracy_score(val_label, val_pred) * 100
        val_auc = metrics.roc_auc_score(val_label, val_score) * 100
        val_ap = metrics.average_precision_score(val_label, val_score) * 100

        print(
            f"Validation loss={val_loss:.4f}, acc={val_acc:.4f}, auc={val_auc:.4f}, ap={val_ap:.4f}"
        )
        wandb.log(
            {
                "val/loss": val_loss,
                "val/acc": val_acc,
                "val/auc": val_auc,
                "val/ap": val_ap,
                "val/epoch": epoch,
            }
        )

    if args.regression:
        ckpt = {
            "net": model.state_dict(),
            "optimizer": optimizer.state_dict(),
            "epoch": epoch,
            "last_val_loss": val_loss,
            "last_val_r2": val_r2,
            "val_score": val_score,
            "val_label": val_label,
        }
        torch.save(ckpt, args.output)
    else:
        ckpt = {
            "net": model.state_dict(),
            "optimizer": optimizer.state_dict(),
            "epoch": epoch,
            "last_val_loss": val_loss,
            "last_val_auc": val_auc,
            "val_score": val_score,
            "val_label": val_label,
        }
        torch.save(ckpt, args.output)
    scheduler.step()

if args.regression:
    wandb.summary.update(
        {
            "last_val_loss": val_loss,
            "last_val_r2": val_r2,
        }
    )
else:
    wandb.summary.update(
        {
            "last_val_loss": val_loss,
            "last_val_auc": val_auc,
        }
    )



import torch
import pandas as pd
import numpy as np
import math
import pickle

from torch.utils.data import Dataset, DataLoader


def load_pickle(f):
    with open(f, "rb") as inFile:
        return pickle.load(inFile)


def split_interval_string(interval_string):
    """chrom:start-end -> (chrom, start, end)"""
    chrom, tmp = interval_string.split(":")
    start, end = map(int, tmp.split("-"))

    return chrom, start, end


class ChromoformerDataset(Dataset):
    def __init__(
        self,
        meta,
        npy_dir,
        target_genes,
        n_feats=7,
        i_max=8,
        binsizes=[2000, 500, 100],
        w_prom=40000,
        w_max=40000,
        regression=False,
    ):
        super(ChromoformerDataset, self).__init__()

        self.npy_dir = npy_dir
        self.n_feats = n_feats
        self.target_genes = target_genes  # List of ENSGs.

        self.meta = pd.read_csv(meta)
        self.regression = regression

        if self.regression:
            self.ensg2label = {
                r.gene_id: np.log2(r.expression + 1) for r in self.meta.to_records()
            }
        else:
            self.ensg2label = {r.gene_id: r.label for r in self.meta.to_records()}

        self.ensg2tss, self.ensg2pcres, self.ensg2scores = {}, {}, {}
        for r in self.meta.to_records():
            self.ensg2tss[r.gene_id] = (r.chrom, r.start, r.end, r.strand)

            if not pd.isnull(r.neighbors):
                self.ensg2pcres[r.gene_id] = r.neighbors.split(";")
                self.ensg2scores[r.gene_id] = [float(s) for s in r.scores.split(";")]
            else:
                self.ensg2pcres[r.gene_id] = []
                self.ensg2scores[r.gene_id] = []

        self.i_max = i_max  # Maximum number of cis-interacting pCREs.
        self.binsizes = binsizes  # List of genomic bin sizes to use.
        self.w_prom = w_prom  # Promoter window size.
        self.w_max = w_max  # Maximum size of pCRE to consider.

    def _bin_and_pad(self, x, bin_size, max_n_bins):
        """Given a 2D tensor x, make binned tensor by
        taking average values of `bin_size` consecutive values.
        Appropriately pad by
        left_pad = ceil((max_n_bins - n_bins) / 2)
        right_pad = floor((max_n_bins - n_bins) / 2)
        """
        l = x.size(1)
        n_bins = math.ceil(l / bin_size)

        # Binning.
        x_binned = []
        for i in range(n_bins):
            b = x[:, i * bin_size : (i + 1) * bin_size].mean(axis=1, keepdims=True)
            b = torch.log(b + 1)
            x_binned.append(b)
        x_binned = torch.cat(x_binned, axis=1)

        # Padding.
        left_pad = math.ceil((max_n_bins - n_bins) / 2)
        right_pad = math.floor((max_n_bins - n_bins) / 2)

        x_binned = torch.cat(
            [
                torch.zeros([x.size(0), left_pad]),
                x_binned,
                torch.zeros([x.size(0), right_pad]),
            ],
            dim=1,
        )

        return x_binned, left_pad, n_bins, right_pad

    def _get_region_representation(
        self, chrom, start, end, bin_size, max_n_bins, strand="+", window=None
    ):
        x = torch.tensor(np.load(f"{self.npy_dir}/{chrom}:{start}-{end}.npy")).float()

        if window is not None:
            x = x[:, 20000 - window // 2 : 20000 + window // 2]
        x, left_pad, n_bins, right_pad = self._bin_and_pad(x, bin_size, max_n_bins)

        if strand == "+":
            return x, left_pad, n_bins, right_pad
        else:
            return torch.fliplr(x), right_pad, n_bins, left_pad

    def __getitem__(self, i):
        target_gene = self.target_genes[i]

        if target_gene not in self.ensg2tss:
            print(target_gene)

        pcres, scores = self.ensg2pcres[target_gene], self.ensg2scores[target_gene]
        n_partners, n_dummies = len(pcres), self.i_max - len(pcres)

        item = {}
        if self.regression:
            item["label"] = torch.tensor(self.ensg2label[target_gene]).float()
        else:
            item["label"] = torch.tensor(self.ensg2label[target_genes]).long()

        item["promoter_feats"] = {}
        item["promoter_pad_masks"] = {}
        item["pcre_feats"] = {}
        item["pcre_pad_masks"] = {}
        item["interaction_masks"] = {}

        chrom_p, start_p, end_p, strand_p = self.ensg2tss[target_gene]
        start_p, end_p = start_p - 20000, start_p + 20000

        for binsize in self.binsizes:
            max_n_bins = self.w_max // binsize

            x_pcres, mask_pcres = [], []

            x_p, left_pad_p, n_bins_p, right_pad_p = self._get_region_representation(
                chrom_p,
                start_p,
                end_p,
                binsize,
                max_n_bins,
                strand_p,
                window=self.w_prom,
            )

            x_p = x_p.permute(1, 0).unsqueeze(0)  # 1 x max_n_bins x n_feats

            mask_p = torch.ones([1, max_n_bins, max_n_bins], dtype=torch.bool)
            mask_p[
                0,
                left_pad_p : left_pad_p + n_bins_p,
                left_pad_p : left_pad_p + n_bins_p,
            ] = 0
            mask_p.unsqueeze_(0)

            interaction_freq = torch.zeros([self.i_max + 1, self.i_max + 1])

            for i, (score, pcre) in enumerate(zip(scores, pcres)):
                chrom_pcre, start_pcre, end_pcre = split_interval_string(pcre)
                if end_pcre - start_pcre > 40000:
                    print(target_gene, chrom_pcre, start_pcre, end_pcre)

                (
                    x_pcre,
                    left_pad_pcre,
                    n_bins_pcre,
                    right_pad_pcre,
                ) = self._get_region_representation(
                    chrom_pcre, start_pcre, end_pcre, binsize, max_n_bins
                )

                mask_pcre = torch.ones([1, max_n_bins, max_n_bins], dtype=torch.bool)
                mask_pcre[
                    0,
                    left_pad_p : left_pad_p + n_bins_p,
                    left_pad_pcre : left_pad_pcre + n_bins_pcre,
                ] = 0

                x_pcres.append(x_pcre)
                mask_pcres.append(mask_pcre)

                interaction_freq[0, i + 1] = score

            x_pcres.append(torch.zeros([self.n_feats, n_dummies * max_n_bins]))
            x_pcres = torch.cat(x_pcres, axis=1).view(-1, self.i_max, max_n_bins)
            x_pcres = x_pcres.permute(1, 2, 0)  # i_max x max_n_bins x 7

            for _ in range(n_dummies):
                m = torch.ones([1, max_n_bins, max_n_bins], dtype=torch.bool)
                mask_pcres.append(m)

            interaction_mask = torch.ones(
                [self.i_max + 1, self.i_max + 1], dtype=torch.bool
            )
            interaction_mask[: n_partners + 1, : n_partners + 1] = 0

            item["promoter_feats"][binsize] = x_p
            item["promoter_pad_masks"][binsize] = mask_p
            item["pcre_feats"][binsize] = x_pcres
            item["pcre_pad_masks"][binsize] = torch.stack(mask_pcres)
            item["interaction_masks"][binsize] = interaction_mask.unsqueeze(0)

        item["interaction_freq"] = interaction_freq
        return item

    def __len__(self):
        return len(self.target_genes)


if __name__ == "__main__":
    import tqdm

    meta = "demo/demo_meta.csv"
    target_genes = pd.read_csv(meta).gene_id.unique()
    npy_dir = "demo/demo_data"

    dataset = ChromoformerDataset(meta, npy_dir, target_genes)
    loader = DataLoader(dataset, batch_size=8, num_workers=1, shuffle=False)

    for i, d in tqdm.tqdm(enumerate(loader), total=len(loader)):
        for binsize in [2000, 500, 100]:
            print(f'{d["promoter_feats"][binsize].shape=}')
            print(f'{d["promoter_pad_masks"][binsize].shape=}')
            print(f'{d["pcre_feats"][binsize].shape=}')
            print(f'{d["pcre_pad_masks"][binsize].shape=}')
            print(f'{d["interaction_masks"][binsize].shape=}')

        print(f'{d["interaction_freq"].shape=}')

        break



import torch
import torch.nn as nn
import torch.nn.functional as F
import math
import numpy as np


class MultiHeadAttention(nn.Module):
    def __init__(self, d_emb, n_heads, d_model, idx, bias=True, scale=True, gate=False):
        super(MultiHeadAttention, self).__init__()

        d_head = d_model // n_heads
        self.n_heads, self.d_head = n_heads, d_head
        self.bias, self.scale = bias, scale
        
        self.gamma_f = nn.Parameter(torch.ones([n_heads]))

        self.w_bias = nn.Linear(2, n_heads, bias=False)
        self.gate = gate
        if gate:
            self.att = nn.Linear(d_emb, 4 * n_heads * d_head, bias=False)  # LinearNoBias for attention
        else:
            self.att = nn.Linear(d_emb, 3 * n_heads * d_head, bias=False)  # LinearNoBias for attention
        self.ff = nn.Linear(n_heads * d_head, d_emb, bias=bias)
        self.ln = nn.LayerNorm(d_emb)
        self.idx = idx

    def forward(self, x, mask=None, bias=None):
        ff_out = self.ff(self._attention(x, mask=mask, bias=bias))
        return self.ln(x + ff_out)

    def _attention(self, x, mask, bias):
        # x : bsz x n_ints x d_model
        bsz, n_ints = x.size(0), x.size(1)

        # self.att(x) : bsz x n_ints x (4 * n_heads * d_head)
        if self.gate:
            wq, wk, wv, gat = torch.chunk(self.att(x), 4, dim=-1) 
            # --> wq, wk, wv : bsz x n_ints x (n_heads * d_head)
            wq, wk, wv, gat = map(lambda x: x.view(bsz, n_ints, self.n_heads, self.d_head), (wq, wk, wv, gat))
            # --> wq, wk, wv : bsz x n_ints x n_heads x d_head
            wq, wk, wv, gat = wq.permute(0, 2, 1, 3), wk.permute(0, 2, 3, 1), wv.permute(0, 2, 1, 3), gat.permute(0, 2, 1, 3)
            # --> wq : bsz x n_heads x n_ints x d_head
            # --> wk : bsz x n_heads x d_head x n_ints
            # --> wv : bsz x n_heads x n_ints x d_head
            # --> gat : bsz x n_heads x n_ints x d_head
        else:
            wq, wk, wv = torch.chunk(self.att(x), 3, dim=-1) 
            # --> wq, wk, wv : bsz x n_ints x (n_heads * d_head)
            wq, wk, wv = map(lambda x: x.view(bsz, n_ints, self.n_heads, self.d_head), (wq, wk, wv))
            # --> wq, wk, wv : bsz x n_ints x n_heads x d_head
            wq, wk, wv = wq.permute(0, 2, 1, 3), wk.permute(0, 2, 3, 1), wv.permute(0, 2, 1, 3)
            # --> wq : bsz x n_heads x n_ints x d_head
            # --> wk : bsz x n_heads x d_head x n_ints
            # --> wv : bsz x n_heads x n_ints x d_head
            # --> gat : bsz x n_heads x n_ints x d_head

        att_score = torch.matmul(wq, wk)
        # --> att_score : bsz x n_heads x n_ints x n_ints
        if self.scale:
            att_score.div_(self.d_head ** 0.5)
        
        if bias is not None:
            # --> bias : bsz x n_ints x n_ints
            bias = bias.unsqueeze(1).expand(bsz, self.n_heads, n_ints, n_ints)
            # --> bias : bsz x n_heads x n_ints x n_ints
            gamma_f = self.gamma_f.unsqueeze(1).unsqueeze(2).expand(self.n_heads, n_ints, n_ints).unsqueeze(0).expand(bsz, self.n_heads, n_ints, n_ints)
            # --> gamma_f = bsz x n_heads x n_ints x n_ints
            att_score += gamma_f * bias

        if mask is not None:
            minus_inf = -65504 if att_score.dtype == torch.float16 else -1e9
            att_score = att_score.masked_fill(mask, minus_inf).type_as(att_score)
        att_prob = F.softmax(att_score, dim=-1)
        # --> att_prob : bsz x n_heads x n_ints x n_ints

        att_vec = torch.matmul(att_prob, wv)
        # --> att_vec : bsz x n_heads x n_ints x d_head

        if self.gate:
            att_vec = att_vec * torch.sigmoid(gat)
            att_vec = att_vec.permute(0, 2, 1, 3).contiguous().contiguous().view(bsz, n_ints, -1)
            # --> att_vec : bsz x n_ints x (h_heads * d_head)
        else:
            att_vec = att_vec.permute(0, 2, 1, 3).contiguous().contiguous().view(bsz, n_ints, -1)
            # --> att_vec : bsz x n_ints x (h_heads * d_head)

        return att_vec


class FeedForward(nn.Module):
    def __init__(self, d_emb, d_ff, activation):
        super(FeedForward, self).__init__()
        self.l1 = nn.Linear(d_emb, d_ff)
        self.l2 = nn.Linear(d_ff, d_emb)
        self.ln = nn.LayerNorm(d_emb)

        self.act = activation

    def forward(self, x):
        return self.ln(x + self.l2(self.act(self.l1(x))))


class AttentionBlock(nn.Module):
    def __init__(self, d_emb, n_heads, d_model, d_ff, idx, activation, gate, bias=True, scale=True):
        super(AttentionBlock, self).__init__()
        self.self_att = MultiHeadAttention(d_emb, n_heads, d_model, idx, bias, scale, gate=gate)
        self.ff = FeedForward(d_emb, d_ff, activation=activation)
    
    def forward(self, x, mask=None, bias=None):
        return self.ff(self.self_att(x, mask, bias))


class Transformer(nn.Module):
    def __init__(self, n_layers, n_heads, d_emb, d_model, d_ff, activation=F.relu, gate=False):
        super(Transformer, self).__init__()
        self.layers = nn.ModuleList([
            AttentionBlock(d_emb, n_heads, d_model, d_ff, idx=i, activation=activation, gate=gate) for i in range(n_layers)
        ])

    def forward(self, x, mask, bias=None):
        for layer in self.layers:
            x = layer(x, mask=mask, bias=bias)
        return x


class PairwiseMultiHeadAttention(nn.Module):
    def __init__(self, d_emb, n_heads, d_model, idx, res_p=0.0, att_p=0.0, bias=True, scale=True, gate=False):
        super(PairwiseMultiHeadAttention, self).__init__()

        d_head = d_model // n_heads
        self.n_heads, self.d_head = n_heads, d_head
        self.bias, self.scale = bias, scale
        
        self.gamma_f = nn.Parameter(torch.ones([n_heads]))
        self.gate = gate

        # Promoter -> Q
        self.p_att = nn.Linear(d_emb, n_heads * d_head, bias=False)
        # pCRE -> K, V
        self.c_att = nn.Linear(d_emb, 2 * n_heads * d_head, bias=False)

        self.ff = nn.Linear(n_heads * d_head, d_emb, bias=bias)

        self.drop_att, self.drop_res = nn.Dropout(att_p), nn.Dropout(res_p)
        self.ln = nn.LayerNorm(d_emb)

        self.idx = idx

    def forward(self, x_p, x_pcre, mask=None, bias=None):
        ff_out = self.ff(self._attention(x_p, x_pcre, mask=mask, bias=bias))
        return self.ln(x_p + self.drop_res(ff_out))

    def _attention(self, x_p, x_pcre, mask, bias):
        # x : bsz x n_ints x d_model
        bsz, n_ints = x_p.size(0), x_p.size(1)
        # self.att(x) : bsz x n_ints x (4 * n_heads * d_head)

        wq = self.p_att(x_p)
        wk, wv = torch.chunk(self.c_att(x_pcre), 2, dim=-1) 
        
        # --> wq, wk, wv : bsz x n_ints x (n_heads * d_head)
        wq, wk, wv = map(lambda x: x.view(bsz, n_ints, self.n_heads, self.d_head), (wq, wk, wv))
        # --> wq, wk, wv : bsz x n_ints x n_heads x d_head
        wq, wk, wv = wq.permute(0, 2, 1, 3), wk.permute(0, 2, 3, 1), wv.permute(0, 2, 1, 3)
        # --> wq : bsz x n_heads x n_ints x d_head
        # --> wk : bsz x n_heads x d_head x n_ints
        # --> wv : bsz x n_heads x n_ints x d_head

        att_score = torch.matmul(wq, wk)
        # --> att_score : bsz x n_heads x n_ints x n_ints
        if self.scale:
            att_score.div_(self.d_head ** 0.5)
        
        if bias is not None:
            # --> bias : bsz x n_ints x n_ints
            bias = bias.unsqueeze(1).expand(bsz, self.n_heads, n_ints, n_ints)
            # --> bias : bsz x n_heads x n_ints x n_ints
            gamma_f = self.gamma_f.unsqueeze(1).unsqueeze(2).expand(self.n_heads, n_ints, n_ints).unsqueeze(0).expand(bsz, self.n_heads, n_ints, n_ints)
            # --> gamma_f = bsz x n_heads x n_ints x n_ints
            att_score += gamma_f * bias

        if mask is not None:
            minus_inf = -65504 if att_score.dtype == torch.float16 else -1e9
            att_score = att_score.masked_fill(mask, minus_inf).type_as(att_score)
        att_prob = self.drop_att(F.softmax(att_score, dim=-1))
        # --> att_prob : bsz x n_heads x n_ints x n_ints

        att_vec = torch.matmul(att_prob, wv)
        # --> att_vec : bsz x n_heads x n_ints x d_head

        att_vec = att_vec.permute(0, 2, 1, 3).contiguous().contiguous().view(bsz, n_ints, -1)
        # --> att_vec : bsz x n_ints x (h_heads * d_head)

        return att_vec


class PairwiseAttentionBlock(nn.Module):
    def __init__(self, d_emb, n_heads, d_model, d_ff, idx, activation, gate, res_p=0.0, att_p=0.0, bias=True, scale=True):
        super(PairwiseAttentionBlock, self).__init__()

        self.self_att = PairwiseMultiHeadAttention(d_emb, n_heads, d_model, idx, res_p, att_p, bias, scale, gate=gate)
        self.ff = FeedForward(d_emb, d_ff, activation=activation)
    
    def forward(self, x_p, x_pcre, mask=None, bias=None):
        return self.ff(self.self_att(x_p, x_pcre, mask, bias))


class PairwiseTransformer(nn.Module):
    def __init__(self, n_layers, n_heads, d_emb, d_model, d_ff, res_p=0.0, att_p=0.0, activation=F.relu, gate=False):
        super(PairwiseTransformer, self).__init__()

        self.layers = nn.ModuleList([
            PairwiseAttentionBlock(d_emb, n_heads, d_model, d_ff, idx=i, res_p=res_p, att_p=att_p, activation=activation, gate=gate) for i in range(n_layers)
        ])

    def forward(self, x_p, x_pcre, mask, bias=None):
        for layer in self.layers:
            x_p = layer(x_p, x_pcre, mask=mask, bias=bias)
        return x_p



import random
import torch
import os
import numpy as np

def seed_everything(seed=42):
    random.seed(seed)
    np.random.seed(seed)
    os.environ["PYTHONHASHSEED"] = str(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)  # type: ignore
    torch.backends.cudnn.deterministic = True  # type: ignore
    torch.backends.cudnn.benchmark = True  # type: ignore


from setuptools import setup, find_packages

setup(
    name="chromoformer",
    packages=find_packages(exclude=[]),
    include_package_data=True,
    version="1.1.2",
    license="MIT",
    description="Chromoformer - Pytorch",
    author="Dohoon Lee",
    author_email="dohlee.bioinfo@gmail.com",
    long_description_content_type="text/markdown",
    url="https://github.com/dohlee/chromoformer",
    keywords=[
        "artificial intelligence",
        "epigenomics",
        "gene expression prediction",
    ],
    install_requires=[
        "einops>=0.3",
        "numpy",
        "torch>=1.6",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.9",
    ],
)
